from distutils.core import setup

setup(
    name = 'dmprofile',
    packages = ['dmprofile'],
    version = 'v0.1.0',
    description = 'pynbody wrapper to facilitate standard analysis of dark matter halos',
    author = 'b-fontana',
    author_email = 'bruno.afonso.alves@ist.utl.pt',
    url = 'https://github.com/b-fontana/DarkMatter',
    download_url = 'https://github.com/b-fontana/DarkMatter/archive/v0.1.0.tar.gz',
    keywords = ['dark matter', 'halos', 'pynbody'],
    classifiers = [],
)
